@@ -17,7 +17,7 @@
 
 import logging
 from collections import deque
-from gevent import sleep
+from gevent import sleep, Timeout as GTimeout
 from gevent.lock import RLock
 
 from .single import SSHClient
@@ -25,7 +25,7 @@ from .tunnel import Tunnel
 from ..common import _validate_pkey_path
 from ..base.parallel import BaseParallelSSHClient
 from ...constants import DEFAULT_RETRIES, RETRY_DELAY
-from ...exceptions import ProxyError, Timeout, HostArgumentException
+from ...exceptions import ProxyError, Timeout, HostArgumentError
 
 
 logger = logging.getLogger(__name__)
@@ -61,21 +61,20 @@ class ParallelSSHClient(BaseParallelSSHClient):
         :param retry_delay: Number of seconds to wait between retries. Defaults
           to :py:class:`pssh.constants.RETRY_DELAY`
         :type retry_delay: int
-        :param timeout: (Optional) SSH session timeout setting in seconds.
-          This controls timeout setting of socket operations used for SSH
-          sessions. Defaults to OS default - usually 60 seconds.
-        :param timeout: (Optional) Individual SSH client timeout setting in
-          seconds passed on to each SSH client spawned by `ParallelSSHClient`.
+        :param timeout: (Optional) Global timeout setting in seconds for all remote
+          operations including all SSH client operations DNS, opening connections,
+          reading output from remote servers,  et al.
 
-          This controls timeout setting of socket operations used for SSH
-          sessions *on a per session basis* meaning for each individual
-          SSH session.
+          For concurrent functions this is a cumulative timeout
+          setting for all concurrent operations. These functions, eg ``run_command`` and
+          ``join``, also allow timeout to be set just for those functions if not
+          set globally via this option.
 
+          For per host operations like
+          ``list(host_out.stdout)`` for reading output it is applied per host if set.
+          Host output read timeout can also be set separately via
+          ``run_command(<..>, read_timeout=<seconds>)``
           Defaults to OS default - usually 60 seconds.
-
-          Parallel functions like `run_command` and `join` have a cummulative
-          timeout setting that is separate to and
-          not affected by `self.timeout`.
         :type timeout: float
         :param pool_size: (Optional) Greenlet pool size. Controls
           concurrency, on how many hosts to execute tasks in parallel.
@@ -114,8 +113,8 @@ class ParallelSSHClient(BaseParallelSSHClient):
           Defaults to False if not set.
           Requires agent forwarding implementation in libssh2 version used.
         :type forward_ssh_agent: bool
-        :param tunnel_timeout: (Optional) Timeout setting for proxy tunnel
-          connections.
+        :param tunnel_timeout: (Optional) Override timeout setting for proxy tunnel
+          connections alone. Defaults to `self.timeout`.
         :type tunnel_timeout: float
 
         :raises: :py:class:`pssh.exceptions.PKeyFileError` on errors finding
@@ -138,13 +137,12 @@ class ParallelSSHClient(BaseParallelSSHClient):
         self._tunnel_in_q = None
         self._tunnel_out_q = None
         self._tunnel_lock = None
-        self._tunnel_timeout = tunnel_timeout
-        self._clients_lock = RLock()
+        self._tunnel_timeout = tunnel_timeout if tunnel_timeout else self.timeout
         self.keepalive_seconds = keepalive_seconds
 
     def run_command(self, command, sudo=False, user=None, stop_on_errors=True,
                     use_pty=False, host_args=None, shell=None,
-                    encoding='utf-8', timeout=None, greenlet_timeout=None,
+                    encoding='utf-8', timeout=None, read_timeout=None,
                     return_list=False):
         """Run command on all hosts in parallel, honoring self.pool_size,
         and return output.
@@ -183,45 +181,36 @@ class ParallelSSHClient(BaseParallelSSHClient):
         :type use_pty: bool
         :param host_args: (Optional) Format command string with per-host
           arguments in ``host_args``. ``host_args`` length must equal length of
-          host list - :py:class:`pssh.exceptions.HostArgumentException` is
+          host list - :py:class:`pssh.exceptions.HostArgumentError` is
           raised otherwise
         :type host_args: tuple or list
         :param encoding: Encoding to use for output. Must be valid
           `Python codec <https://docs.python.org/library/codecs.html>`_
         :type encoding: str
-        :param timeout: (Optional) Timeout in seconds for reading from stdout
-          or stderr. Defaults to no timeout. Reading from stdout/stderr will
+        :param read_timeout: (Optional) Timeout in seconds for reading from stdout
+          or stderr. Reading from stdout/stderr will
           raise :py:class:`pssh.exceptions.Timeout`
-          after ``timeout`` number seconds if remote output is not ready.
-        :type timeout: int
-        :param greenlet_timeout: (Optional) Greenlet timeout setting.
-          Defaults to no timeout. If set, this function will raise
-          :py:class:`gevent.Timeout` after ``greenlet_timeout`` seconds
-          if no result is available from greenlets.
-          In some cases, such as when using proxy hosts, connection timeout
-          is controlled by proxy server and getting result from greenlets may
-          hang indefinitely if remote server is unavailable. Use this setting
-          to avoid blocking in such circumstances.
-          Note that ``gevent.Timeout`` is a special class that inherits from
-          ``BaseException`` and thus **can not be caught** by
-          ``stop_on_errors=False``.
-        :type greenlet_timeout: float
-        :param return_list: (Optional) Return a list of ``HostOutput`` objects
-          instead of dictionary. ``run_command`` will return a list starting
-          from 2.0.0 - enable this flag to avoid client code breaking on
-          upgrading to 2.0.0.
+          after ``timeout`` seconds when set if remote output is not ready.
+        :type read_timeout: float
+        :param timeout: Deprecated - use read_timeout. Same as
+          read_timeout and kept for backwards compatibility - to be removed
+          in future release.
+        :type timeout: float
+        :param return_list: No-op - list of ``HostOutput`` always returned.
+          Parameter kept for backwards compatibility - to be removed in future
+          releases.
         :type return_list: bool
         :rtype: Dictionary with host as key and
           :py:class:`pssh.output.HostOutput` as value
           *or* list(:py:class:`pssh.output.HostOutput`) when
           ``return_list=True``
-        :raises: :py:class:`pssh.exceptions.AuthenticationException` on
+        :raises: :py:class:`pssh.exceptions.AuthenticationError` on
           authentication error
-        :raises: :py:class:`pssh.exceptions.UnknownHostException` on DNS
+        :raises: :py:class:`pssh.exceptions.UnknownHostError` on DNS
           resolution error
-        :raises: :py:class:`pssh.exceptions.ConnectionErrorException` on error
+        :raises: :py:class:`pssh.exceptions.ConnectionError` on error
           connecting
-        :raises: :py:class:`pssh.exceptions.HostArgumentException` on number of
+        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
           host arguments not equal to number of hosts
         :raises: :py:class:`TypeError` on not enough host arguments for cmd
           string format
@@ -238,13 +227,13 @@ class ParallelSSHClient(BaseParallelSSHClient):
         return BaseParallelSSHClient.run_command(
             self, command, stop_on_errors=stop_on_errors, host_args=host_args,
             user=user, shell=shell, sudo=sudo,
-            encoding=encoding, use_pty=use_pty, timeout=timeout,
-            greenlet_timeout=greenlet_timeout, return_list=return_list)
+            encoding=encoding, use_pty=use_pty,
+            return_list=return_list, read_timeout=read_timeout if read_timeout else timeout,
+        )
 
     def __del__(self):
         if not hasattr(self, '_host_clients'):
             return
-        logger.debug("Disconnecting clients")
         for s_client in self._host_clients.values():
             try:
                 s_client.disconnect()
@@ -279,46 +268,38 @@ class ParallelSSHClient(BaseParallelSSHClient):
         auth_thread_pool = True
         if self.proxy_host is not None and self._tunnel is None:
             self._start_tunnel_thread()
-        logger.debug("Make client request for host %s, host in clients: %s",
-                     host, host in self.host_clients)
-        with self._clients_lock:
-            if (host_i, host) not in self._host_clients \
-               or self._host_clients[(host_i, host)] is None:
-                _user, _port, _password, _pkey = self._get_host_config_values(
-                    host)
-                proxy_host = None if self.proxy_host is None else '127.0.0.1'
-                if proxy_host is not None:
-                    auth_thread_pool = False
-                    _wait = 0.0
-                    max_wait = self.timeout if self.timeout is not None else 60
-                    with self._tunnel_lock:
-                        self._tunnel_in_q.append((host, _port))
+        logger.debug("Make client request for host %s, (host_i, host) in clients: %s",
+                     host, (host_i, host) in self._host_clients)
+        if (host_i, host) not in self._host_clients \
+           or self._host_clients[(host_i, host)] is None:
+            _user, _port, _password, _pkey = self._get_host_config_values(host_i, host)
+            proxy_host = None if self.proxy_host is None else '127.0.0.1'
+            if proxy_host is not None:
+                auth_thread_pool = False
+                max_wait = self.timeout if self.timeout is not None else 60
+                with self._tunnel_lock:
+                    self._tunnel_in_q.append((host, _port))
+                with GTimeout(seconds=max_wait, exception=Timeout):
                     while True:
-                        if _wait >= max_wait:
-                            raise Timeout("Timed out waiting on tunnel to "
-                                          "open listening port")
                         try:
                             _port = self._tunnel_out_q.pop()
                         except IndexError:
-                            logger.debug(
-                                "Waiting on tunnel to open listening port")
-                            sleep(.5)
-                            _wait += .5
+                            logger.debug("Waiting on tunnel to open listening port")
+                            sleep(.1)
                         else:
                             break
-                _client = SSHClient(
-                    host, user=_user, password=_password, port=_port,
-                    pkey=_pkey, num_retries=self.num_retries,
-                    timeout=self.timeout,
-                    allow_agent=self.allow_agent, retry_delay=self.retry_delay,
-                    proxy_host=proxy_host, _auth_thread_pool=auth_thread_pool,
-                    forward_ssh_agent=self.forward_ssh_agent,
-                    keepalive_seconds=self.keepalive_seconds,
-                    identity_auth=self.identity_auth,
-                )
-                self.host_clients[host] = _client
-                self._host_clients[(host_i, host)] = _client
-                return _client
+            _client = SSHClient(
+                host, user=_user, password=_password, port=_port,
+                pkey=_pkey, num_retries=self.num_retries,
+                timeout=self.timeout,
+                allow_agent=self.allow_agent, retry_delay=self.retry_delay,
+                proxy_host=proxy_host, _auth_thread_pool=auth_thread_pool,
+                forward_ssh_agent=self.forward_ssh_agent,
+                keepalive_seconds=self.keepalive_seconds,
+                identity_auth=self.identity_auth,
+            )
+            self._host_clients[(host_i, host)] = _client
+            return _client
         return self._host_clients[(host_i, host)]
 
     def copy_file(self, local_file, remote_file, recurse=False, copy_args=None):
@@ -347,7 +328,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
         :param copy_args: (Optional) format local_file and remote_file strings
           with per-host arguments in ``copy_args``. ``copy_args`` length must
           equal length of host list -
-          :py:class:`pssh.exceptions.HostArgumentException` is raised otherwise
+          :py:class:`pssh.exceptions.HostArgumentError` is raised otherwise
         :type copy_args: tuple or list
 
         :rtype: list(:py:class:`gevent.Greenlet`) of greenlets for remote copy
@@ -355,7 +336,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
 
         :raises: :py:class:`ValueError` when a directory is supplied to
           local_file and recurse is not set
-        :raises: :py:class:`pssh.exceptions.HostArgumentException` on number of
+        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
           per-host copy arguments not equal to number of hosts
         :raises: :py:class:`pss.exceptions.SFTPError` on SFTP initialisation
           errors
@@ -412,7 +393,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
         :param copy_args: (Optional) format remote_file and local_file strings
           with per-host arguments in ``copy_args``.   ``copy_args`` length must
           equal length of host list -
-          :py:class:`pssh.exceptions.HostArgumentException` is raised otherwise
+          :py:class:`pssh.exceptions.HostArgumentError` is raised otherwise
         :type copy_args: tuple or list
         :param encoding: Encoding to use for file paths.
         :type encoding: str
@@ -422,7 +403,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
 
         :raises: :py:class:`ValueError` when a directory is supplied to
           local_file and recurse is not set
-        :raises: :py:class:`pssh.exceptions.HostArgumentException` on number of
+        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
           per-host copy arguments not equal to number of hosts
         :raises: :py:class:`pss.exceptions.SFTPError` on SFTP initialisation
           errors
@@ -456,7 +437,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
             self._host_clients[(host_i, host)].scp_recv, host,
             remote_file, local_file, recurse=recurse)
 
-    def scp_send(self, local_file, remote_file, recurse=False):
+    def scp_send(self, local_file, remote_file, recurse=False, copy_args=None):
         """Copy local file to remote file in parallel via SCP.
 
         This function returns a list of greenlets which can be
@@ -488,9 +469,22 @@ class ParallelSSHClient(BaseParallelSSHClient):
         :raises: :py:class:`pss.exceptions.SCPError` on errors copying file.
         :raises: :py:class:`OSError` on local OS errors like permission denied.
         """
-        return [self.pool.spawn(self._scp_send, host_i, host, local_file,
-                                remote_file, recurse=recurse)
-                for host_i, host in enumerate(self.hosts)]
+        copy_args = [{'local_file': local_file,
+                      'remote_file': remote_file}
+                     for i, host in enumerate(self.hosts)] \
+            if copy_args is None else copy_args
+        local_file = "%(local_file)s"
+        remote_file = "%(remote_file)s"
+        try:
+            return [self.pool.spawn(self._scp_send, host_i, host,
+                                    local_file % copy_args[host_i],
+                                    remote_file % copy_args[host_i],
+                                    recurse=recurse)
+                    for host_i, host in enumerate(self.hosts)]
+        except IndexError:
+            raise HostArgumentError(
+                "Number of per-host copy arguments provided does not match "
+                "number of hosts")
 
     def scp_recv(self, remote_file, local_file, recurse=False, copy_args=None,
                  suffix_separator='_'):
@@ -534,7 +528,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
         :param copy_args: (Optional) format remote_file and local_file strings
           with per-host arguments in ``copy_args``. ``copy_args`` length *must*
           equal length of host list -
-          :py:class:`pssh.exceptions.HostArgumentException` is raised otherwise
+          :py:class:`pssh.exceptions.HostArgumentError` is raised otherwise
         :type copy_args: tuple or list
 
         :rtype: list(:py:class:`gevent.Greenlet`) of greenlets for remote copy
@@ -542,7 +536,7 @@ class ParallelSSHClient(BaseParallelSSHClient):
 
         :raises: :py:class:`ValueError` when a directory is supplied to
           local_file and recurse is not set.
-        :raises: :py:class:`pssh.exceptions.HostArgumentException` on number of
+        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
           per-host copy arguments not equal to number of hosts.
         :raises: :py:class:`pss.exceptions.SCPError` on errors copying file.
         :raises: :py:class:`OSError` on local OS errors like permission denied.
@@ -569,6 +563,6 @@ class ParallelSSHClient(BaseParallelSSHClient):
                 local_file % copy_args[host_i], recurse=recurse)
                     for host_i, host in enumerate(self.hosts)]
         except IndexError:
-            raise HostArgumentException(
+            raise HostArgumentError(
                 "Number of per-host copy arguments provided does not match "
                 "number of hosts")
